from collections import defaultdict
import os
from pathlib import Path
import subprocess

import click
import importlib.metadata
from typing import List, Literal

from loguru import logger

from pbt.config import PBTConfig
from pbt.diff import RemoteDiff
from pbt.package import search_packages, topological_sort, update_versions
from pbt.git import Git
from pbt.pypi import PyPI


@click.group()
@click.version_option(importlib.metadata.version("polyrepo-bt"))
def cli():
    pass


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the package that we want to build. If empty, build all packages",
)
@click.option(
    "-i",
    "--install",
    is_flag=True,
    help="Install other dependencies of the package (not the inter/local dependencies)",
)
@click.option(
    "-e",
    "--editable",
    is_flag=True,
    help="Whether to install dependencies in editable mode",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Whether to force reinstall the dependencies",
)
@click.option("--cwd", default="", help="Override current working directory")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="increase verbosity",
)
def make(
    package: List[str],
    install: bool = False,
    editable: bool = False,
    force: bool = False,
    cwd: str = "",
    verbose: bool = False,
):
    """Make package"""
    pbt_cfg = PBTConfig.from_dir(cwd)
    packages = search_packages(pbt_cfg)

    if len(package) == 0:
        make_packages = set(packages.keys())
    else:
        make_packages = set(package)

    if len(make_packages.difference(packages.keys())) > 0:
        raise Exception(
            f"Passing unknown packages: {make_packages.difference(packages.keys())}. Available options: {list(packages.keys())}"
        )

    # mapping from package name to whether the content has been changed since the last built
    built_pkgs = {}
    updated_version_pkgs = set()
    for pkg_name in make_packages:
        pkg = packages[pkg_name]
        if install:
            logger.debug("Install {} external dependencies", pkg_name)
            pkg.install(without_inter_dependency=True, verbose=verbose)

        dep_pkgs = pkg.all_inter_dependencies()
        update_versions(set(dep_pkgs.keys()).difference(updated_version_pkgs), packages)
        updated_version_pkgs = updated_version_pkgs.union(dep_pkgs.keys())

        for dep_name in topological_sort(dep_pkgs):
            if dep_name not in built_pkgs:
                # TODO: optimize this code as we don't need to rebuild if we are install in editable mode
                built_pkgs[dep_name] = dep_pkgs[dep_name].build(pbt_cfg)
                if built_pkgs[dep_name] and dep_name in pkg.dependencies:
                    # update if there is a change in the dependency make it no longer compatible,
                    if not pkg.is_package_compatible(dep_pkgs[dep_name]):
                        pkg.update_package_version(dep_pkgs[dep_name])
            if built_pkgs[dep_name] or force:
                pkg.install_dep(
                    dep_pkgs[dep_name],
                    pbt_cfg,
                    editable=editable,
                    no_build=True,
                    verbose=verbose,
                )
    return


@click.command()
@click.option("--cwd", default="", help="Override current working directory")
def update(cwd: str = ""):
    pbt_cfg = PBTConfig.from_dir(cwd)
    packages = search_packages(pbt_cfg)
    update_versions(list(packages.keys()), packages, force=True)


@click.command()
@click.option(
    "-p",
    "--package",
    multiple=True,
    help="Specify the package that we want to build. If empty, build all packages",
)
@click.option("--cwd", default="", help="Override current working directory")
def publish(package: str, cwd: str = ""):
    pbt_cfg = PBTConfig.from_dir(cwd)
    packages = search_packages(pbt_cfg)

    if len(package) == 0:
        publish_packages = set(packages.keys())
    else:
        publish_packages = set(package)

    if len(publish_packages.difference(packages.keys())) > 0:
        raise Exception(
            f"Passing unknown packages: {publish_packages.difference(packages.keys())}. Available options: {list(packages.keys())}"
        )

    all_pub_pkgs = {}
    for pkg_name in publish_packages:
        pkg = packages[pkg_name]
        dep_pkgs = pkg.all_inter_dependencies()

        all_pub_pkgs[pkg.name] = pkg
        all_pub_pkgs.update(dep_pkgs)

    update_versions(all_pub_pkgs.keys(), packages)
    pypi = PyPI.get_instance()
    has_error = False

    all_pub_pkgs = [
        all_pub_pkgs[pkg_name] for pkg_name in topological_sort(all_pub_pkgs)
    ]
    pkg2diff = {}

    for pkg in all_pub_pkgs:
        remote_pkg_version, remote_pkg_hash = pypi.get_latest_version_and_hash(
            pkg.name
        ) or (None, None)
        diff = RemoteDiff.from_pkg(pkg, pbt_cfg, remote_pkg_version, remote_pkg_hash)
        if not diff.is_version_diff and diff.is_content_changed:
            logger.error(
                "Package {} has been modified, but its version hasn't been updated",
                pkg.name,
            )
            has_error = True
        pkg2diff[pkg.name] = diff

    if has_error:
        raise Exception(
            "Stop publishing because some packages have been modified but their versions haven't been updated. Please see the logs for more information"
        )

    for pkg in all_pub_pkgs:
        if pkg2diff[pkg.name].is_version_diff:
            logger.info("Publish package {}", pkg.name)
            pkg.publish()


@click.command()
@click.option(
    "--repo", default="", help="Specify the poly-repository that we are working with."
)
@click.option("--cwd", default="", help="Override current working directory")
@click.argument("subcommand")
def git(repo: str, cwd: str, subcommand: Literal["snapshot"]):
    """Execute Git commands in a super-project"""
    if subcommand == "clone":
        assert repo.endswith(".git"), f"Invalid repository: `{repo}`"

        # clone repository
        repo_dir = Git.clone_all(repo, cwd)

        # checkout the submodule to the correct branch
        for submodule in Git.find_submodules(repo_dir):
            commit_id = Git.get_current_commit(submodule)
            branches = Git.get_branches_contain_commit(submodule, commit_id)

            # select branch to checkout to, prefer development branch as we want to
            # resume previous work
            names = {branch.get_name() for branch in branches}
            if len(names.difference(["master", "main"])) > 0:
                name = list(names.difference(["master", "main"]))[0]
            else:
                name = "master" if "master" in names else "main"

            brs = [branch for branch in branches if branch.get_name() == name]
            has_local = any(br.is_local() for br in brs)
            if has_local:
                Git.checkout_branch(submodule, name)
            else:
                Git.checkout_branch(submodule, name, create=True)
    elif subcommand == "snapshot":
        pbt_cfg = PBTConfig.from_dir(cwd)
        cwd = str(pbt_cfg.cwd.absolute())

        for submodule_dir in Git.find_submodules(cwd):
            # get the current branch
            branch = Git.get_current_branch(submodule_dir)
            print(f"bash -c 'cd {submodule_dir}; git checkout {branch}; git pull'")
    else:
        raise Exception(f"Invalid subcommand: {subcommand}")


cli.add_command(make)
cli.add_command(publish)
cli.add_command(update)
cli.add_command(git)

if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pbt']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.5.3,<0.6.0',
 'orjson>=3.6.3,<4.0.0',
 'poetry>=1.1.8,<2.0.0',
 'pytest-mock>=3.6.1,<4.0.0',
 'rocksdb>=0.7.0,<0.8.0',
 'semver>=2.13.0,<3.0.0',
 'toml>=0.10.2,<0.11.0']

extras_require = \
{':python_version < "3.9"': ['graphlib_backport>=1.0.0,<2.0.0']}

entry_points = \
{'console_scripts': ['pbt = pbt.cli:cli']}

setup_kwargs = {
    'name': 'polyrepo-bt',
    'version': '1.7.0',
    'description': 'A build tool for poetry packages (living in submodules of a Git repo)',
    'long_description': None,
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import re
import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, List, Literal, Optional

import semver
from loguru import logger

if TYPE_CHECKING:
    from pbt.package import Package

from functools import cached_property


class Poetry:
    def __init__(self, package: "Package"):
        self.package = package

    @cached_property
    def env_path(self) -> Path:
        """Get environment path of the package, create it if doesn't exist"""
        output = (
            subprocess.check_output(
                ["poetry", "env", "list", "--full-path"], cwd=self.package.dir
            )
            .decode()
            .strip()
        )

        if output == "":
            # environment doesn't exist, create it
            temp_output = (
                subprocess.check_output(
                    ["poetry", "run", "python", "-m", "__hello__"], cwd=self.package.dir
                )
                .decode()
                .strip()
            )
            assert temp_output.endswith("Hello world!")
            output = (
                subprocess.check_output(
                    ["poetry", "env", "list", "--full-path"], cwd=self.package.dir
                )
                .decode()
                .strip()
            )

        lines = output.split("\n")
        if len(lines) > 1:
            logger.warning(
                "There are multiple virtual environments for package {}, and we pick the first one (maybe incorrect)",
                self.package.name,
            )

        if lines[0].endswith(" (Activated)"):
            path = lines[0].split(" ")[0]
        else:
            path = lines[0]

        assert os.path.exists(path)
        return Path(path)

    @cached_property
    def pip_path(self) -> Path:
        return self.env_path / "bin/pip"

    @cached_property
    def python_path(self) -> Path:
        return self.env_path / "bin/python"

    def install(self, without_inter_dependency: bool = True, verbose: bool = False):
        if verbose:
            call = subprocess.check_call
        else:
            call = subprocess.check_output

        if without_inter_dependency:
            with self.temporary_mask(
                [dep_pkg.name for dep_pkg in self.package.inter_dependencies]
            ):
                call(["poetry", "install"], cwd=self.package.dir)
        else:
            call(["poetry", "install"], cwd=self.package.dir)

    def publish(self):
        subprocess.check_output(["poetry", "publish", "--build"], cwd=self.package.dir)

    def destroy(self):
        """Remove the virtual environment"""
        try:
            subprocess.check_output(
                ["poetry", "env", "remove", "python"], cwd=self.package.dir
            )
        except subprocess.CalledProcessError as e:
            pass

    def update_version(self, rule: Literal["major", "minor", "patch"]):
        subprocess.check_output(["poetry", "version", rule], cwd=self.package.dir)

    def replace_version(self, version: Optional[str] = None):
        version = version or self.package.version
        with open(os.path.join(self.package.dir, "pyproject.toml"), "r") as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                if re.match("version *= *", line) is not None:
                    lines[i] = f'version = "{version}"\n'
                    break
            else:
                raise Exception(
                    "Can not find the version of the package in pyproject.toml"
                )

        with open(os.path.join(self.package.dir, "pyproject.toml"), "w") as f:
            for line in lines:
                f.write(line)

    @classmethod
    def is_version_compatible(cls, version: str, constraint: str) -> bool:
        """Check if version is compatible"""
        m = re.match(
            r"(?P<op>[\^\~]?)(?P<major>\d+)\.((?P<minor>\d+)\.(?P<patch>\d+)?)?",
            constraint,
        )
        assert m is not None, "The constraint is too complicated to handle for now"

        lowerbound = semver.VersionInfo(
            major=int(m.group("major")),
            minor=int(m.group("minor") or "0"),
            patch=int(m.group("patch") or "0"),
        )
        if m.group("op") == "^":
            # special case for 0 following the nodejs way (I can't believe why)
            # see more: https://nodesource.com/blog/semver-tilde-and-caret/
            if lowerbound.major == 0:
                if lowerbound.minor == 0:
                    upperbound = lowerbound.bump_patch()
                else:
                    upperbound = lowerbound.bump_minor()
            else:
                upperbound = lowerbound.bump_major()
        elif m.group("op") == "~":
            if m.group("patch") is not None:
                upperbound = lowerbound.bump_minor()
            elif m.group("minor") is not None:
                upperbound = lowerbound.bump_minor()
            else:
                upperbound = lowerbound.bump_major()
        else:
            upperbound = lowerbound.bump_patch()
        return lowerbound <= version < upperbound

    def update_inter_dependency(self, pkg_name: str, pkg_version: str):
        should_update = False
        with open(os.path.join(self.package.dir, "pyproject.toml"), "r") as f:
            lines = f.readlines()
            match_lines = [
                (i, line)
                for i, line in enumerate(lines)
                if re.match(f"{pkg_name} *= *", line) is not None
            ]
            assert len(match_lines) == 1
            idx, line = match_lines[0]
            m = re.match(
                rf"""({pkg_name} *= *)(?:['"])([^0-9]*)([^'"]+)(?:['"])(.*)""",
                line,
                flags=re.DOTALL,
            )
            if m is not None:
                groups = m.groups()
                prev_version = groups[2]
                if prev_version != pkg_version:
                    should_update = True
                    logger.info(
                        f"In {self.package.name}, bump {pkg_name} from `{prev_version}` to `{pkg_version}`"
                    )
                    lines[idx] = f'{groups[0]}"{groups[1]}{pkg_version}"{groups[3]}'
            else:
                raise NotImplementedError(f"Do not know how to parse `{line}` yet")

        if should_update:
            with open(os.path.join(self.package.dir, "pyproject.toml"), "w") as f:
                for line in lines:
                    f.write(line)

    @contextmanager
    def temporary_mask(self, deps: List[str]):
        """Temporary mask out selected dependencies of the package. This is usually used for installing the package"""
        with open(self.package.dir / "pyproject.toml", "r") as f:
            lines = f.readlines()
        for i, line in enumerate(lines):
            if any([re.match(f"{dep} *=", line) is not None for dep in deps]):
                lines[i] = "# " + lines[i]
        try:
            os.rename(
                self.package.dir / "pyproject.toml",
                self.package.dir / "pyproject.toml.backup",
            )
            with open(self.package.dir / "pyproject.toml", "w") as f:
                f.write("".join(lines))
            yield None
        finally:
            os.rename(
                self.package.dir / "pyproject.toml.backup",
                self.package.dir / "pyproject.toml",
            )



import glob
import os
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union, NamedTuple


class GitFileStatus(NamedTuple):
    is_deleted: bool
    fpath: str


class GitBranch(NamedTuple):
    local: Optional[str]
    remote: Optional[str]
    is_active: bool

    def get_name(self) -> str:
        if self.local is not None:
            return self.local.rsplit("/", 1)[-1]
        assert self.remote is not None
        return self.remote.rsplit("/", 1)[-1]

    def is_local(self) -> bool:
        return self.local is not None and self.local.find("/") == -1


PathOrStr = Union[str, Path]


class Git:
    @classmethod
    def get_new_modified_deleted_files(cls, cwd: PathOrStr):
        git_dir = (
            subprocess.check_output(["git", "rev-parse", "--show-toplevel"], cwd=cwd)
            .decode()
            .strip()
        )
        output = subprocess.check_output(
            ["git", "status", "-uall", "--porcelain=v1", "--no-renames", "."], cwd=cwd
        ).decode()
        # rstrip as the first line can have empty character
        output = output.rstrip()
        if len(output) == 0:
            return []

        lines = output.split("\n")
        results = []

        # TODO: take a look at this document and implement it correctly
        #  https://git-scm.com/docs/git-status
        for line in lines:
            code = line[:2]
            rel_file_path = line[3:].strip()
            assert " -> " not in rel_file_path and "R" not in code
            results.append(
                GitFileStatus(
                    is_deleted="D" in code, fpath=os.path.join(git_dir, rel_file_path)
                )
            )
        return results

    @classmethod
    def get_current_commit(cls, cwd: PathOrStr):
        return (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode()
            .strip()
        )

    @classmethod
    def get_current_branch(cls, cwd: PathOrStr):
        return (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"], cwd=cwd
            )
            .decode()
            .strip()
        )

    @classmethod
    def checkout_branch(cls, cwd: PathOrStr, branch: str, create: bool = False):
        if not create:
            # exist local branch
            cmd = ["git", "checkout", branch]
        else:
            cmd = ["git", "checkout", "-b", branch]
        subprocess.check_call(cmd, cwd=cwd)

    @classmethod
    def get_branches_contain_commit(cls, cwd: PathOrStr, commit_id: str):
        output = subprocess.check_output(
            ["git", "branch", "-a", "--contains", commit_id], cwd=cwd
        ).decode()
        branches = []

        for x in output.split("\n"):
            x = x.strip()
            if x == "":
                continue

            is_active = False
            if x[0] == "*":
                is_active = True
                x = x[2:]

            if x.startswith("(HEAD detached"):
                # detached head, so no branch
                continue

            if x.find(" -> ") != -1:
                remote, local = x.split(" -> ")
            elif x.startswith("remotes/"):
                remote = x
                local = None
            else:
                remote = None
                local = x

            branch = GitBranch(local=local, remote=remote, is_active=is_active)
            if branch.get_name() == "HEAD":
                continue
            branches.append(branch)

        return branches

    @classmethod
    def init(cls, cwd: PathOrStr):
        subprocess.check_output(["git", "init"], cwd=cwd)

    @classmethod
    def commit_all(cls, cwd: PathOrStr, msg: str = "add all files"):
        subprocess.check_output(["git", "add", "-A"], cwd=cwd)
        subprocess.check_output(["git", "commit", "-m", f"'{msg}'"], cwd=cwd)

    @classmethod
    def clone_all(cls, repo: str, cwd: PathOrStr) -> Path:
        """Clone a repository and its submodules. Then, return the cloned directory"""
        # get repo name to create directory to clone the repo into
        repo_name = repo.rsplit("/", 1)[-1].replace(".git", "")
        repo_dir = Path(cwd) / repo_name
        if repo_dir.exists():
            raise Exception(
                f"Directory {repo_dir} already exists. Can't clone the repository"
            )
        if not repo_dir.parent.exists():
            raise Exception(
                f"Directory {repo_dir.parent} does not exist. Can't clone the repository into it"
            )

        subprocess.check_call(
            [
                "git",
                "clone",
                "--recurse-submodules",
                "-j8",
                repo,
            ],
            cwd=cwd,
        )

        return repo_dir

    @classmethod
    def find_submodules(cls, repo_dir: PathOrStr) -> List[Path]:
        submodules = []
        repo_dir = os.path.abspath(repo_dir)

        for dir in Path(repo_dir).iterdir():
            if not dir.is_dir():
                continue
            superdir = (
                subprocess.check_output(
                    ["git", "rev-parse", "--show-superproject-working-tree"],
                    cwd=str(dir),
                )
                .decode()
                .strip()
            )
            # only consider submodules that are of the current project
            if superdir == repo_dir:
                submodules.append(dir)

        return submodules


if __name__ == "__main__":
    res = Git.get_new_modified_deleted_files("/workspace/sm-dev/osin")
    print("\n".join([str(x) for x in res]))
